# -*- coding: utf-8 -*-
"""
Created on Sun Feb 17 16:36:47 2019

@author: yoelr
"""
from ... import Stream, Species, HeatUtility, Unit
from . import Facility
from ..decorators import cost

__all__ = ('BoilerTurbogenerator',)

lps = HeatUtility.heating_agents['Low pressure steam']

#: TODO add reference of NREL

@cost('Work', 'Turbogenerator',
      CE=551, S=-42200, kW=-42200, cost=9500e3, n=0.83, BM=1.8)
@cost('Flow rate', 'Hot process water softener system', 
      CE=551, cost=78e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Amine addition pkg', 
      CE=551, cost=40e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Deaerator',
      CE=551, cost=305e3, S=235803, n=0.6, BM=3.0)
@cost('Flow rate', 'Boiler',
      CE=551, cost=28550e3, kW=2752, S=238686, n=0.6, BM=1.8)
class BoilerTurbogenerator(Facility):
    """Create a Boiler_TurboGenerator object that will calculate electricity generation from burning the feed. It also takes into account how much steam is being produced, and the required cooling utility of the turbo generator. No emissions or mass balances are taken into account.
    
    **Parameters**
    
        **boiler_efficiency:** [float] Fraction of heat transfered to steam.
    
        **turbo_generator_efficiency:** [float] Fraction of steam heat converted to electricity.
    
    **ins**
    
        [0] Feed that will be burned.
    
    **outs**
    
        [0] Emission (burned feed)
        
    
    """
    _N_heat_utilities = 2
    _has_power_utility = True
    _units = {'Flow rate': 'kg/hr',
              'Work': 'kW'}
    
    def __init__(self, ID='', ins=None, outs=(), *,
                 boiler_efficiency=0.80,
                 turbo_generator_efficiency=0.85):
        Unit.__init__(self, ID, ins, outs)
        self.boiler_efficiency = boiler_efficiency
        self.turbo_generator_efficiency = turbo_generator_efficiency
        self.steam_utilities = set()
        self.steam_demand = Stream(None, species=Species('Water'), P=lps.P, T=lps.T, phase='g')
    
    def _run(self): pass
    
    def _design(self):
        B_eff = self.boiler_efficiency
        TG_eff = self.turbo_generator_efficiency
        steam = self.steam_demand
        steam_utilities = self.steam_utilities
        if not steam_utilities:
            for u in self.system._costunits:
                if u is self: continue
                for hu in u._heat_utilities:
                    if hu.ID == 'Low pressure steam':
                        steam_utilities.add(hu)
        steam._mol[0] = sum(i.flow for i in steam_utilities)
        feed = self._ins[0]
        emission = self._outs[0]
        hu_cooling, hu_steam = self._heat_utilities
        H_steam = steam.H
        feed_Hc = feed.Hc
        
        # This is simply for the mass balance (no special purpose)
        emission.mol[:] = feed.mol # TODO: In reality, this should be CO2
        
        Design = self._Design
        Design['Flow rate'] = feed_massnet = feed.massnet
        
        # 50 percent of bagasse is water, so remove latent heat of vaporization
        H_content = feed_Hc*B_eff - feed_massnet*1130 
        
        # Heat available for the turbogenerator
        H_electricity = H_content - H_steam 
        
        #: [float] Total steam produced by the boiler (kmol/hr)
        self.total_steam = H_content/lps.Hvap 
        
        if H_electricity < 0:
            H_steam = H_content
            cooling = electricity = 0
        else:
            electricity = H_electricity * TG_eff
            cooling = electricity - H_electricity
        hu_cooling(cooling, steam.T)
        hu_steam.ID = 'Low pressure steam'
        molnet = steam.molnet
        hu_steam.cost = -(molnet*lps.price_kmol + H_steam*lps.price_kJ)
        Design['Work'] = (-electricity/3600)        

# Simulation of ethanol production from sugarcane
# in Brazil: economic study of an autonomous
# distillery
# Marina O.S. Diasa,b, Marcelo P. Cunhaa, Charles D.F. Jesusa, Mirna I.G.
# Scandiffioa, Carlos E.V. Rossella,b, Rubens Maciel Filhob, Antonio Bonomia
# a CTBE – Bioethanol Science and Technology National Laboratory, PO Box 6170 –
# CEP 13083-970, Campinas – SP, Brazil, marina.dias@bioetanol.org.br
# b School of Chemical Engineering, University of Campinas, PO Box 6066 – CEP
# 13083-970, Campinas – SP, Brazil
# LHV = 7565 # Bagasse lower heating value (kJ/kg)


# -*- coding: utf-8 -*-
"""
Created on Wed May  1 19:05:53 2019

@author: yoelr
"""
import biosteam as bst
import copy
from ._design import design
from math import ceil

__all__ = ('cost', 'add_cost', 'CostItem')

class CostItem:
    """Create a CostItem object which defines exponential scaling factors for an item's purchase cost.
    
    **Parameters**
    
        **basis:** Name of size parameter used for scaling.
    
        **units:** Units of measure.
        
        **S:** Size.
        
        **ub:** Size limit.
        
        **CE:** Chemical engineering plant cost index.
        
        **cost:** Purchase cost of item.
        
        **n:** Exponential factor.
        
        **kW:** Electricity rate.
        
        **BM:** Bare module factor (installation factor).
    
    """
    __slots__ = ('_basis', '_units', 'S', 'ub', 'CE', 'cost', 'n', 'kW', 'BM')
    def __init__(self, basis, units, S, ub, CE, cost, n, kW, BM):
        self._basis = basis
        self._units = units
        self.S = S
        self.ub = ub
        self.CE = CE
        self.cost = cost
        self.n = n
        self.kW = kW
        self.BM = BM
    
    __getitem__ = object.__getattribute__
    __setitem__ = object.__setattr__
    
    @property
    def basis(self):
        return self._basis
    @property
    def units(self):
        return self._units
    
    def __repr__(self):
        return f"<{type(self).__name__}: {self._basis} ({self._units})>"
    
    def _ipython_display_(self):
        print(f"{type(self).__name__}: {self._basis} ({self._units})\n"
             +f" S     {self.S:.3g}\n"
             +f" ub    {self.ub:.3g}\n"
             +f" CE    {self.CE:.3g}\n"
             +f" cost  {self.cost:.3g}\n"
             +f" n     {self.n:.3g}\n"
             +f" kW    {self.kW:.3g}\n"
             +f" BM    {self.BM:.3g}")
    show = _ipython_display_

def _cost(self):
    D = self._Design
    C = self._Cost
    kW = 0
    for i, x in self.cost_items.items():
        S = D[x._basis]
        if x.ub:
            D['#'+i] = N = ceil(S/x.ub)
            q = S/x.S
            F = q/N
            C[i] = N*bst.CE/x.CE*x.cost*F**x.n
            kW += x.kW*q
        else:
            F = S/x.S
            C[i] = bst.CE/x.CE*x.cost*F**x.n
            kW += x.kW*F
    if kW: self._power_utility(kW)

def _extended_cost(self):
    _cost(self)
    self._end_decorated_cost_

@property
def BM(self):
    raise AttributeError("can only get installation factor 'BM' through 'cost_items'")
@BM.setter 
def BM(self, BM):
    raise AttributeError("can only set installation factor 'BM' through 'cost_items'")
BM_property = BM
del BM

@property
def installation_cost(self):
    C = self._Cost
    return ([C[i]*j.BM for i,j in self.cost_items.items()]).sum()

def cost(basis, ID=None, *, CE, cost, n, S=1, ub=0, kW=0, BM=1, units=None, fsize=None):    
    r"""Add item (free-on-board) purchase cost based on exponential scale up:
    
    **Parameters**
    
        **basis:** Name of size parameter used for scaling.
        
        **ID:** Name of purchase item.
        
        **CE:** Chemical engineering plant cost index.
        
        **cost:** Purchase cost of item.
        
        **n:** Exponential factor.
        
        **S:** Size.
        
        **ub:** Size limit.
        
        **kW:** Electricity rate.
        
        **BM:** Bare module factor (installation factor).
        
        **units:** Units of measure.
        
        **fsize:** Function that accepts a Unit object argument and returns the size parameter. If None, defaults to function predefined for given name and units of measure.
        
    **Examples**
    
        :doc:`Unit decorators`
    
    """
    
    return lambda cls: add_cost(cls, ID, basis, units, S, ub, CE, cost, n, kW, BM, fsize)

def add_cost(cls, ID, basis, units, S, ub, CE, cost, n, kW, BM, fsize):
    if kW: cls._has_power_utility = True
    if basis in cls._units:
        if fsize:
            raise RuntimeError(f"cost basis '{basis}' already defined in class, cannot pass 'fsize' argument")
        elif not units: 
            units = cls._units[basis]
        elif units != cls._units[basis]:
            raise RuntimeError(f"cost basis '{basis}' already defined in class with units '{cls._units[basis]}'")
    elif units:
        design._add_design2cls(cls, basis, units, fsize)
    else:
        raise RuntimeError(f"units of cost basis '{basis}' not available in '{cls.__name__}._units' dictionary, must pass units or define in class")
    if hasattr(cls, 'cost_items'):
        if 'cost_items' not in cls.__dict__:
            cls.cost_items = copy.deepcopy(cls.cost_items)
        if not ID:
            raise ValueError("must pass an 'ID' for purchase cost item")
        if ID in cls.cost_items:
            raise ValueError(f"ID '{ID}' already in use")
        cls.cost_items[ID] = CostItem(basis, units, S, ub, CE, cost, n, kW, BM)
    else:
        if '_cost' in cls.__dict__:
            raise RuntimeError("'_cost' method is already implemented, cannot decorate class")
        ID = ID or cls.line
        cls.cost_items = {ID: CostItem(basis, units, S, ub, CE, cost, n, kW, BM)}
        if hasattr(cls, '_end_decorated_cost_'):
            cls._cost = _extended_cost
        else:
            cls._cost = _cost
        cls.BM = BM_property
        cls.installation_cost = installation_cost
    return cls




# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:45:47 2018

@author: yoelr
"""
import numpy as np
from ._hx import HXutility
from .. import Stream, Unit
from scipy.integrate import odeint
from .decorators import cost
from ._tank import MixTank
from .designtools import size_batch
from ..reaction import Reaction

@cost('Reactor volume', 'Cleaning in place', CE=521.9,
      cost=421e3, S=3785, n=0.6, BM=1.8)
@cost('Reactor volume', 'Agitators', CE=521.9, cost=52500,
      S=3785, n=0.5, kW=22.371, BM=1.5)
@cost('Reactor volume', 'Reactors', CE=521.9, cost=844000,
      S=3785, n=0.5, BM=1.5)
class Fermentation(Unit):
    """Create a Fermentation object which models large-scale batch fermentation for the production of 1st generation ethanol using yeast [1, 2, 3, 4]. Only sucrose and glucose are taken into account. Conversion is based on reaction time, `tau`. Cleaning and unloading time, `tau_0`, fraction of working volume, `V_wf`, and number of reactors, `N_reactors`, are attributes that can be changed. Cost of a reactor is based on the NREL batch fermentation tank cost assuming volumetric scaling with a 6/10th exponent [3].
    
    **Parameters**
    
        **tau:** Reaction time.
        
        **efficiency:** User enforced efficiency.
        
        **N:** Number of batch reactors
    
    **ins**
    
        [:] Inffluent streams
        
    **outs**
    
        [0] Effluent
        
        [1] CO2
    
    **References**
    
        [1] Oliveira, Samuel C., et al. "Discrimination between ethanol inhibition models in a continuous alcoholic fermentation process using flocculating yeast." Applied biochemistry and biotechnology 74.3 (1998): 161-172.
        
        [2] Oliveira, Samuel C., et al. "Continuous ethanol fermentation in a tower reactor with flocculating yeast recycle: scale-up effects on process performance, kinetic parameters and model predictions." Bioprocess Engineering 20.6 (1999): 525-530.
        
        [3] Oliveira, Samuel C., et al. "Mathematical modeling of a continuous alcoholic fermentation process in a two-stage tower reactor cascade with flocculating yeast recycle." Bioprocess and biosystems engineering 38.3 (2015): 469-479.
        
        [4] Oliveira, Samuel C., et al. "Kinetic Modeling of 1‐G Ethanol Fermentations." Fermentation Processes. InTech, 2017.
        
        [5] D. Humbird, R. Davis, L. Tao, C. Kinchin, D. Hsu, and A. Aden National. Renewable Energy Laboratory Golden, Colorado. P. Schoen, J. Lukas, B. Olthof, M. Worley, D. Sexton, and D. Dudgeon. Harris Group Inc. Seattle, Washington and Atlanta, Georgia. Process Design and Economics for Biochemical Conversion of Lignocellulosic Biomass to Ethanol Dilute-Acid Pretreatment and Enzymatic Hydrolysis of Corn Stover. May 2011. Technical Report NREL/TP-5100-47764
        
    .. note::
        
        A compound with CAS 'Yeast' must be present in species.
        
    **Examples**
    
        :doc:`Fermentation Example`
        
    """
    _units = {'Reactor volume': 'm3',
              'Cycle time': 'hr',
              'Loading time': 'hr',
              'Total dead time': 'hr'}
    
    _N_heat_utilities = 0
    _has_power_utility = True
    line = 'Fermentation'    
    
    #: Cleaning and unloading time (hr)
    tau_0 = 3
    
    #: Fraction of filled tank to total tank volume
    working_volume_fraction = MixTank.working_volume_fraction
    _V_wf = 0.9
    
    #: tuple of kinetic parameters for the kinetic model. Default constants are fitted for Oliveria's model (mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a)
    kinetic_constants = (0.31,  # mu_m1
                         1.01,  # mu_m2
                         1.88,  # Ks1
                         2.81,  # Ks2
                         82.8,  # Pm1
                         108.2, # Pm2
                         113.4, # Xm
                         0.45,  # Y_PS
                         0.18)  # a
    
    
    def __init__(self, ID='', ins=None, outs=(), *, 
                 tau,  N, efficiency=0.9, iskinetic=False):
        Unit.__init__(self, ID, ins, outs)
        self.hydrolysis = Reaction('Sucrose + Water -> 2Glucose', 'Sucrose', 1.00)
        self.fermentation = Reaction('Glucose -> 2Ethanol + 2CO2',  'Glucose', efficiency)
        self.iskinetic = iskinetic
        self.efficiency = efficiency
        self.tau = tau
        self.N = N
        self._cooler = hx = HXutility(None)
        self._heat_utilities = hx._heat_utilities
        hx._ins = hx._outs
        vent, effluent = self.outs
        hx._outs[0].T = effluent.T = vent.T = 305.15
        vent.phase = 'g'

    def _calc_efficiency(self, feed, tau):
        # Get initial concentrations
        y, e, s, w = feed.indices(['Yeast',
                                   '64-17-5',
                                   '492-61-5',
                                   '7732-18-5'])
        mass = feed.mass
        volnet = feed.volnet
        concentration_in = mass/volnet
        X0, P0, S0 = (concentration_in[i] for i in (y, e, s))
        
        # Integrate to get final concentration
        t = np.linspace(0, tau, 1000)
        C_t = odeint(self.kinetic_model, (X0, P0, S0), t,
                     args=self.kinetic_constants)
        # Cache data
        self._X = C_t[:, 0]
        self._P = C_t[:, 1]
        self._S = S = C_t[:, 2]
        
        # Calculate efficiency
        Sf = S[-1]
        Sf = Sf if Sf > 0 else 0
        Y_PS = self.kinetic_constants[-2]
        eff = (S0 - Sf)/S0 * Y_PS/0.511
        return eff
        
    @staticmethod
    def kinetic_model(z, t, *kinetic_constants) -> '(dXdt, dPdt, dSdt)':
        """Return change of yeast, ethanol, and substrate concentration in kg/m3.
        
        **Parameters**
        
            **z:** Iterable with (X, E, S) [-]:
                * X: Yeast concentration (kg/m3)
                * P: Ethanol concentration (kg/m3)
                * S: Substrate concentration (kg/m3)
            
            **t:** Time point
            
            ***kinetic_constants:**
                * mu_m1: Maximum specific growth rate (1/hr)
                * mu_m2: Maximum specific ethanol production rate (g-product/g-cell-hr)
                * Ks1: Sugar saturation constant for growth (g/L)
                * Ks2: Sugar saturation constant for product (g/L)
                * Pm1: Maximum product concentration at zero growth [mu_m1=0] (g/L)
                * Pm2: Maximum product concentration [mu_m2=0] (g/L)
                * Xm: Maximum cell concentration [mu_m1=0] (g/L)
                * Y_PS: Ethanol yield based on sugar consumed
                * a: Toxic power
                
        """
        mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a = kinetic_constants
        
        # Current yeast, ethanol, and glucose concentration (kg/m3)
        X, P, S = z
        
        # Compute coefficients
        mu_X = mu_m1 * (S/(Ks1 + S)) * (1 - P/Pm1)**a*((1-X/Xm))
        mu_P = mu_m2 * (S/(Ks2 + S)) * (1 - P/Pm2)
        mu_S = mu_P/0.45
        
        # Compute derivatives
        dXdt = mu_X * X
        dPdt = (mu_P * X)
        dSdt =  - mu_S * X
        return (dXdt, dPdt, dSdt)
       
    @property
    def N(self):
        """[int] Number of reactors"""
        return self._N
    @N.setter
    def N(self, N):
        if N <= 1:
            raise ValueError(f"number of reactors must be greater than 1, value {N} is infeasible")
        self._N = N

    @property
    def efficiency(self):
        return self.fermentation.X
    @efficiency.setter
    def efficiency(self, efficiency):
        self.fermentation.X = efficiency

    @property
    def tau(self):
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau

    def _run(self):
        vent, effluent = self.outs
        Stream.sum(effluent, self.ins)
        effluent_mol = effluent.mol
        self.hydrolysis(effluent_mol)
        if self.iskinetic:
            self.fermentation.X = self._calc_efficiency(effluent, self._tau)
        self.fermentation(effluent_mol)
        vent.copyflow(effluent, ('CO2',), remove=True)
        y = effluent.P_vapor/101325
        Y = y.sum()
        mol2vent = vent.molnet*Y/(1-Y)*y
        effluent_mol[:] -= mol2vent
        vent.mol += mol2vent
    
    def number_of_reactors_at_minimum_cost(self):
        cost_old = np.inf
        self._N, N = 2, self._N
        self.simulate()
        cost_new = self.purchase_cost
        while cost_new < cost_old:
            self._N += 1
            self.simulate()
            cost_old = cost_new
            cost_new = self.purchase_cost
        self._N, N = N, self._N
        return N - 1
        
    def _design(self):
        v_0 = self.outs[1].volnet
        tau = self._tau
        tau_0 = self.tau_0
        Design = self._Design
        Design['Number of reactors'] = N = self._N
        Design.update(size_batch(v_0, tau, tau_0, N, self._V_wf))
        hx = self._cooler
        hx.outs[0]._mol[:] = self.outs[0].mol/N 
        hu = hx._heat_utilities[0]
        hu(self._Hnet/N, self.outs[0].T)
        hx._design(hu.duty)
        hx._cost()
        hu.duty *= N
        hu.cost *= N
        hu.flow *= N
    
_cost = Fermentation._cost
def _extended_cost(self):
    _cost(self)
    N = self._Design['Number of reactors']
    Cost = self._Cost
    Cost['Coolers'] = self._cooler._Cost['Heat exchanger']
    for i in Cost: Cost[i] *= N
Fermentation._cost = _extended_cost


# -*- coding: utf-8 -*-
"""
Created on Sun Jul  7 20:26:38 2019

@author: yoelr
"""
import sys
import importlib
from types import ModuleType

__all__ = ('LazyPkg',)

class LazyPkg(ModuleType):
    
    def __init__(self, __name__, modules):
        super().__init__(__name__)
        dct = self.__dict__
        dct.update(sys.modules[__name__].__dict__)
        sys.modules[__name__] = self
        try:
            __all__ = self.__all__
            for i in modules:
                if i in __all__:
                    raise ValueError(f"module '{i}' already in {__name__}.__all__")
            __all__.extend(modules)
        except:
            raise ValueError("__all__ must be present in the module as a list")
    
    def __dir__(self):
        return self.__all__
    
    def __getattr__(self, name):
        if name in self.__all__:
            module = importlib.import_module('.'+name, self.__package__)
            setattr(self, name, module)
            return module
        else:
            for i in self.__all__:
                module = getattr(self, i)
                if isinstance(module, ModuleType) and hasattr(module, name):
                    attr = getattr(module, name)
                    setattr(self, name, attr)
                    return attr
        raise AttributeError(f"module {self.__name__} has not attribute '{name}'")
        
        


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 20 18:40:05 2019

@author: yoelr
"""

__all__ = []

from ._VLEsolver import *
from .unifac import *

from . import unifac
from . import _VLEsolver

__all__.extend(unifac.__all__)
__all__.extend(_VLEsolver.__all__)


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 17 18:43:39 2019

@author: yoelr
"""
from . import Facility
import biosteam as bst

__all__ = ('ProcessWaterCenter',)

cost = bst.units.decorators.cost

@cost('Makeup water flow rate', 'Makeup water pump',
      CE=551, kW=20*0.7457, cost=6864, S=155564, n=0.8, BM=3.1)
@cost('Process water flow rate', 'Process water pump',
      CE=551, kW=75*0.7457, cost=15292, S=518924, n=0.8, BM=3.1)
@cost('Process water flow rate', 'Tank',
      CE=522, cost=250e3, S=451555, n=0.7, BM=1.7)
class ProcessWaterCenter(Facility):
    _N_ins = 2
    _N_outs = 1
    makeup_water_price = 0.0002113 #: (USD/kg)
    boiler_blowdown = 0.002
    RO_rejection = 0.25
    CT_evaporation = 0.001
    CT_blowdown = 0.01
    _units = {'Makeup water flow rate': 'kg/hr',
              'Process water flow rate': 'kg/hr'}
    def __init__(self, ID, ins=None, outs=(), *, BT, CT):
        Facility.__init__(self, ID, ins, outs)
        makeup_water = bst.Stream('Makeup_water', species=bst.Species('Water',)) 
        if self._ins[1]:
            self._ins[1].link = makeup_water
        else:
            self._ins[1] = makeup_water
        self.BT = BT
        self.CT = CT
        
    def _run(self):
        process_water = self._outs[0].molnet
        recycle_water = self._ins[0].molnet
        boiler = self.BT.total_steam * self.boiler_blowdown * 1/(1-self.RO_rejection)
        cooling_tower = self.CT.cooling_water * (self.CT_evaporation + self.CT_blowdown)
        makeup_water_stream = self._ins[1] 
        Design = self._Design
        process_loss = process_water - recycle_water
        if process_loss > 0:
            makeup_water = boiler + cooling_tower + process_loss
        else:
            makeup_water = boiler + cooling_tower
        makeup_water_stream.mol[0] = makeup_water
        
        Design['Process water flow rate'] = (process_water + boiler + cooling_tower)*18.015
        Design['Makeup water flow rate'] = makeup_water*18.015
        makeup_water_stream.price = self.makeup_water_price
        
        


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 00:35:01 2019

@author: yoelr
"""
from .._exceptions import SolverError
#from .other_utils import count
import numpy as np

__all__ = ('bounded_secant', 'secant', 'wegstein', 'conditional_wegstein')

def secant(f, x0, xtol, args=(), maxiter=50):
    x1 = x0*(1 + 1e-4) + (-1e-4 if x0 < 0 else 1e-4)
    pos = abs
    y0 = f(x0, *args)
    y1 = f(x1, *args) 
    for iter in range(maxiter): 
        x_diff = x1-x0 
        x1 = x0 - y1*(x_diff)/(y1-y0)
        if pos(x_diff) < xtol: return x1
        x0 = x1
        y0 = y1
        y1 = f(x1, *args)
    raise SolverError(f'failed to converge after {maxiter} iterations')

def bounded_secant(f, x0, x, x1, y0, y1, xtol, args=()):
    assert y0*y1 < 0, ('objective function bounderies '
                       'must have opposite signs')
    if y1 < 0: x0, y0, x1, y1 = x1, y1, x0, y0
    y = f(x, *args)
    x_ = x0
    while abs(x-x_) > xtol:
        if y > 0:
            x_ = x1 = x
            y1 = y
        elif y < 0:
            x_ = x0 = x
            y0 = y
        else: return x
        x = x0 - y0*(x1-x0)/(y1-y0)
        y = f(x, *args)
    return x

# def accelerated_secant(f, x0, xtol, ytol, args=(), maxiter=50):
#     x1 = x0*(1 + 1e-4) + (-1e-4 if x0 < 0 else 1e-4)
#     pos = abs
#     y0 = f(x0, *args)
#     if pos(y0) < ytol: return x0
#     y1 = f(x1, *args)
#     if pos(y1) < ytol: return x1 
#     x1 = g0 = x0 - y1*(x1-x0)/(y1-y0)
#     if (pos(x1-x0) < xtol): return x1
#     g1 = x0
#     y0 = y1
#     for iter in range(maxiter): 
#         y1 = f(x1, *args)
#         g1 = x1 - y1*(x1-x0)/(y1-y0)
#         if (pos(g1-x1) < xtol) or (pos(y1) < ytol): return g1
#         w = (x1-x0)/(x1-g1 + g0-x0)
#         x1 = w*g1 + (1-w)*x1
#         x0 = x1
#         y0 = y1
#         g0 = g1
#     raise SolverError(f'failed to converge after {maxiter} iterations')

def wegstein(f, x0, xtol, args=(), maxiter=50):
    len_ = len(x0)
    ones = np.ones(len_)
    s = np.zeros(len_)
    x1 = y0 = f(x0, *args)
    for iter in range(maxiter):
        if (abs(y0 - x0) < xtol).all(): return x1
        y1 = f(x1, *args)
        x_diff = x1 - x0
        pos = x_diff != 0
        s[pos] = (y1[pos] - y0[pos])/x_diff[pos]
        x0 = x1
        y0 = y1
        s[s > 0.9] = 0.9
        w = ones/(ones-s)
        x1 = w*y1 + (1-w)*x1
    raise SolverError(f'failed to converge after {maxiter} iterations')

def conditional_wegstein(f, x0):
    len_ = len(x0)
    ones = np.ones(len_)
    s = np.zeros(len_)
    y0, condition = f(x0)
    x1 = y0
    while condition:
        y1, condition = f(x1)
        x_diff = x1 - x0
        pos = x_diff != 0
        s[pos] = (y1[pos] - y0[pos])/x_diff[pos]
        x0 = x1
        y0 = y1
        s[s > 0.9] = 0.9
        w = ones/(ones-s)
        x1 = w*y1 + (1-w)*x1
    return y1


    

#    if y0*y1 < 0:
#             x = x0 - y1*(x1-x0)/(y1-y0)
#             return bounded_secant(f, x0, x, x1, y0, y1, xtol, ytol, args, maxiter-iter)


